//========================================================================================
//
// WORDBOX - Generate random pronounceable words based on trigram probabilities
//
// WordBox is Copyright (c) 1999 by Klaus Rindtorff
//
// Permission to use, copy, and distribute this software and its documentation
// for any purpose, without fee, and without a written agreement is hereby
// granted, provided that the above copyright notice and this paragraph and
// the following two paragraphs appear in all copies.
//
// IN NO EVENT SHALL THE AUTHOR BE LIABLE TO ANY PARTY FOR DIRECT, INDIRECT,
// SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES, INCLUDING LOST PROFITS,
// ARISING OUT OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE
// AUTHOR HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
//
// THE AUTHOR SPECIFICALLY DISCLAIMS ANY WARRANTIES, INCLUDING, BUT NOT
// LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A
// PARTICULAR PURPOSE. THE SOFTWARE PROVIDED HEREUNDER IS ON AN "AS IS"
// BASIS, AND THE AUTHOR HAS NO OBLIGATIONS TO PROVIDE MAINTENANCE, SUPPORT,
// UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
//
// WORDBOX generates random pronounceable words based on trigram probabilities.
// Each word is generated randomly based on the probability of individual trigrams in
// a given language. The probabilities for each language are stored in a database.
//
// Each database consists of 26 tables for the starting character 'A'..'Z'. Each table
// contains 26 rows for the middle character, and each row contains 28 entries for the
// third character 'A'..'Z' and two extra values. Each entry gives the probability for
// the occurence of the respective trigram. The first extra value gives the probability
// that the relevant character pair is a word prefix. The second extra value gives the
// probability that the character pair is a word suffix. All probabilities are relative
// and all values are normalized to be in the range 0 to 255.
//
// The trigram statistics were generated from word lists for the following languages:
//
//   ENGLISH          153418 words
// Version 1.00  1998-12-4   Klaus Rindtorff, klaus@ivorytower.com
// Version 1.10  1999-01-19  minor modifications, added comments, added Names database
// Version 1.20  1999-07-11  changes for Waba Beta8 and wextras 1.01
//                           modified wextras List class to compile with Waba Beta8
// Version 1.30  1999-12-28  changes for Waba Beta9 and wextras 1.21
//                           added Japanese names
//
//========================================================================================
import java.util.Date;
import java.net.URL;
import palm.*;

public class WordBox extends Object
{

  PDBFile catalog;
  byte[] table = new byte[26*28];
  int prefixSum;
  int randomSeed;
  int record = -1;
  URL cb;

  public WordBox(URL codebase)
  {
    Date t = new Date(); // initialize the random number generator
    randomSeed = (int)(t.getTime()/1000);
		cb=codebase;
  }


  private int prefixSum()
  {
    int i,j,sum=0;
    for (i=0; i<26; i++) for (j=0; j<26; j++) sum += probability(i,j,26);
    return sum;
  }

  private int infixSum(int i, int j)
  {
    int k,sum=0;
    for (k=0; k<26; k++) sum += probability(i,j,k);
    return sum;
  }

  private int suffixSum(int i, int j)
  {
    int k,sum=0;
    for (k=0; k<26; k++) if (probability(i,j,k)!=0) sum += probability(j,k,27);
    return sum;
  }

  private int random(int range) // range may be 1..65536
  {
    randomSeed = ((randomSeed * 32145) + 1) & 0x0FFFF;
    return (randomSeed * range) >>> 16;
  }

  private int probability(int i, int j, int k)
  {
  	int nbr;

    if (record != i) // cache table
    {
    	try {
	      catalog.setRecordPos(i);
				catalog.readBytes(table,0,table.length);
    	  record = i;
      }
      catch (Exception e) {
      	System.out.println("Exception="+e);
      	System.out.println("Exception reading row="+i);
        return 0;
      }
    }
    int value = table[j*28+k];
    return (value<0)? value+256 : value; // adjust byte value
  }

  private boolean setDatabase() {
      if (catalog == null) {
	      try {
        	//System.out.println("H="+cb.getHost()+" F="+cb.getFile()+" P="+cb.getProtocol()+" T="+cb.getPort());
  	    	//System.out.println(cb.getFile()+"english.pdb");
	  	    catalog = new PDBFile(cb,"english.pdb",PDBFile.READ_ONLY);
      	  prefixSum = prefixSum(); // calculate only once
				}
  	    catch (Exception e) {
    	  	System.out.println("Open Exception="+e);
	    	  return false; // not found
	      }
      }
    return true;
  }

  public String generate(int wordLength) {
    int i,j,k,count,length,sum,sigma;
    StringBuffer word;

    if (!setDatabase())
    	return "No Database";

    word = new StringBuffer();
    length = 0;

    while (length < wordLength) {
      select_prefix:
      while (true) { // select a word prefix (2 characters)
        sigma = random(prefixSum);
        sum = 0;
        i = 0;
        while (i < 26) {
          j = 0;
          while (j < 26) {
            sum += probability(i,j,26);
            if (sum > sigma) {
              if (length++ < wordLength) word.append((char)('A'+i));
              if (length++ < wordLength) word.append((char)('A'+j));
              break select_prefix;
            }
 	          j++;
   	      }
     	    i++;
       	}
       }
 	    select_next:
   	  while (length < wordLength) { // select more characters
     	  if (length == wordLength-1 && length > 1) { // select a suffix
       	  sigma = suffixSum(i,j);
         	if (sigma != 0) { // is there a possible continuation?
           	sigma = random(sigma);
            sum = 0;
 	          for (k=0; k<26; k++) {
   	          // does this make an existing trigram?
     	        if (probability(i,j,k) != 0) sum += probability(j,k,27);
       	      if (sum > sigma) {
         	      word.append((char)('A'+k));
           	    length++;
             	  break select_next;
              }
 	          }
   	      }
     	  }
       	sigma = infixSum(i,j);
        if (sigma == 0) break select_next; // is this a dead-end?
        sigma = random(sigma);
 	 	    sum = 0;
     	  select_trigram:
       	for (k=0; k<26; k++) {
         	sum += probability(i,j,k);
	        if (sum > sigma) {
            word.append((char)('A'+k));
   	        length++;
     	      i=j; j=k;
       	    break select_trigram;
         	}
	      }
      }
   	}
	  return word.toString();
  }
}

